"""
This is a class that offers a set of functions to work with meteorological data
in ecsv or grib(2) format.

Created by Pere Munar-Adrover
email: pere.munaradrover@gmail.com
Further developed and mainted by
Mykhailo Dalchenko (mykhailo.dalchenko@unige.ch) and
Georgios Voutsinas (georgios.voutsinas@unige.ch)
"""

import os
import sys
import logging
import json
import itertools

import astropy.units as u
from astropy.table import Table, QTable, vstack, Column
import numpy as np
from scipy.interpolate import interp1d
from bisect import bisect_left

from molecularprofiles.utils.grib_utils import get_grib_file_data, extend_grib_data
from molecularprofiles.utils.constants import (
    DENSITY_SCALE_HEIGHT,
    N0_AIR,
    STD_GRAVITATIONAL_ACCELERATION,
    STD_AIR_DENSITY,
    STD_CORSIKA_ALTITUDE_PROFILE,
    RAYLEIGH_SCATTERING_ALTITUDE_BINS,
)

from molecularprofiles.utils.humidity import (
    compressibility,
    density_moist_air,
    molar_fraction_water_vapor,
    partial_pressure_water_vapor,
)
from molecularprofiles.utils.rayleigh import Rayleigh

ROOTDIR = os.path.dirname(os.path.abspath(__file__))
log_config_file = f"{ROOTDIR}/utils/mdps_log.conf"
logging.config.fileConfig(fname=log_config_file, disable_existing_loggers=False)
logger = logging.getLogger(__name__)


class MolecularProfile:
    """
    This class provides a series of functions to analyze the quality of the data for both
    CTA-North and CTA-South.

    Methods within this class:

    get_data:                   it retrieves the data from the input file. If the input file
                                is a grib file and there is no file in the working directory
                                with the same name but with .txt extension the program extracts
                                the data from the grib file through the grib_utils program. If
                                there is such a txt file, it reads it directly
    write_atmospheric_profile:  prints the data into a txt file which format is compliant with the input card
                                for the CORSIKA air shower simulation software
    create_mdp:                 creates an altitude profile of the molecular number density
    rayleigh_extinction:        creates a file, in format to be directly fed to simtel simulation, with the
                                extinction per altitude bin for wavelengths from 200 to 1000 nm
    """

    def __init__(self, data_file):
        """
        Constructor

        :param data_file: txt file containing the data (string)
        """

        self.data_file = data_file

    # =============================================================================================================
    # Private functions
    # =============================================================================================================
    def _interpolate_cubic(self, x_param, y_param, new_x_param):
        func = interp1d(x_param, y_param, kind="cubic", bounds_error=False)
        return func(new_x_param)

    def _compute_mass_density(self, air="moist", co2_concentration=415):
        """
        Computes regular and exponential mass density of air.

        Adds to data the following columns:
        - 'Xw': molar fraction of water vapor (0 if air is dry)
        - 'Compressibility'
        - 'Mass Density'
        - 'Exponential Mass Density'

        Parameters
        ----------
        air : str
            Type of air, can be 'moist' or 'dry'
        co2_concentration : float
            CO2 volume concentration in ppmv
        """

        if air == "moist":
            self.data["Xw"] = molar_fraction_water_vapor(
                self.data["Pressure"], self.data["Temperature"], self.data["Relative humidity"]
            )
        elif air == "dry":
            self.data["Xw"] = 0.0
        else:
            raise ValueError("Wrong air condition. It must be 'moist' or 'dry'.")

        self.data["Compressibility"] = compressibility(
            self.data["Pressure"], self.data["Temperature"], self.data["Xw"]
        )
        self.data["Mass Density"] = density_moist_air(
            self.data["Pressure"],
            self.data["Temperature"],
            self.data["Compressibility"],
            self.data["Xw"],
            co2_concentration,
        )
        self.data["Exponential Mass Density"] = (
            self.data["Mass Density"] / STD_AIR_DENSITY
        ).decompose() * np.exp((self.data["Altitude"] / DENSITY_SCALE_HEIGHT).decompose())

    # =============================================================================================================
    # Main get data function
    # =============================================================================================================
    def get_data(self):
        """
        Reads ECMWF or GDAS data in ecsv or grib(2) format
        and computes statistical description of the data
        """
        file_ext = os.path.splitext(self.data_file)[1]
        if file_ext == ".grib" or file_ext == ".grib2":
            self.data = get_grib_file_data(self.data_file)
            self.data = extend_grib_data(self.data)
        elif file_ext == ".ecsv":
            self.data = Table.read(self.data_file)
        else:
            raise NotImplementedError(
                f"Only grib (1,2) and ecsv formats are supported at the moment. Requested format: {file_ext}"
            )
        self.stat_columns = [
            "Pressure",
            "Altitude",
            "Density",
            "Temperature",
            "Wind Speed",
            "Wind Direction",
            "Relative humidity",
            "Exponential Density",
        ]
        self.stat_data = self.data[self.stat_columns].group_by("Pressure")
        self.stat_description = {
            "avg": self.stat_data.groups.aggregate(np.mean),
            "std": self.stat_data.groups.aggregate(np.std),
            "mad": self.stat_data.groups.aggregate(lambda x: np.mean(np.absolute(x - np.mean(x)))),
            "p2p_max": self.stat_data.groups.aggregate(lambda x: np.max(x) - np.mean(x)),
            "p2p_min": self.stat_data.groups.aggregate(lambda x: np.mean(x) - np.min(x)),
        }

    def _refractive_index(self, P, T, RH, wavelength, C):
        """Wrapper for Rayleigh.calculate_n()."""
        rayleigh = Rayleigh(wavelength, C, P, T, RH)
        return rayleigh.refractive_index

    def _take_closest(self, myList, myNumber):
        """
        Returns closest value to myNumber.
        If two numbers are equally close, return the smallest number.
        """
        pos = bisect_left(myList, myNumber)
        if pos == 0:
            return myList[0]
        if pos == len(myList):
            return myList[-1]
        before = myList[pos - 1]
        after = myList[pos]
        if after - myNumber < myNumber - before:
            return after
        else:
            return before

    def _pick_up_reference_atmosphere(self, ceiling, floor, reference_atmosphere):
        """
        Picks up the reference atmosphere corresponding to the season and
        the geographical location. It selects all rows above the given ceiling.

        Parameters:
        -----------
        ceiling
            Astropy quantity expressing the ceiling of the DAS data.
        reference_atmosphere
            ecsv file with the reference atmosphere profile.

        Returns:
        --------
        table
            Astropy table with the atmospheric profile above the given ceiling.
        """
        try:
            reference_atmosphere_table = Table.read(reference_atmosphere)
        except Exception as ex:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(ex).__name__, ex.args)
            logger.error(message)
            sys.exit(1)
        try:
            mask = (
                reference_atmosphere_table["altitude"]
                >= ceiling.to(reference_atmosphere_table["altitude"].unit)
            ) | (
                reference_atmosphere_table["altitude"]
                < floor.to(reference_atmosphere_table["altitude"].unit)
            )
            return reference_atmosphere_table[mask]
        except Exception as ex:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(ex).__name__, ex.args)
            logger.error(message)
            sys.exit(1)

    # =======================================================================================================
    # printing functions:
    # =======================================================================================================

    def write_atmospheric_profile(
        self,
        outfile,
        co2_concentration,
        floor=None,
        reference_atmosphere=None,
        altitude_list=STD_CORSIKA_ALTITUDE_PROFILE,
    ):
        """
        Write an output file in the style of a CORSIKA atmospheric configuration file:
        altitude (km)     density (g/cm^3)   thickness (g/cm^2)   refractive index -1   temperature (K)   pressure (hPA)   partial water pressure

        Parameters:
        -----------
        outfile : string
            Name of the returned file.
        co2_concentration : float
            12MACOBAC value
        floor : Quantity
            Lowest altitude level where the atmospheric parameters will be calculated. If none is provided, it gets the closer to the sea level altitude where data are available. Units of length.
        reference_atmosphere : path
            The path where the file with the reference atmosphere model is located.
        altitude_list : Quantity
            Tuple with the altitudes that the atmospheric parameters will be calculated. Units of length.

        Returns:
        --------
            Ecsv file in the style of a CORSIKA atmospheric configuration file.
        """

        if floor is None:
            floor = (self.stat_description["avg"]["Altitude"][-1]) * self.stat_description["avg"][
                "Altitude"
            ].unit
        ceiling = self._take_closest(
            altitude_list.to(self.stat_description["avg"]["Altitude"].unit),
            (self.stat_description["avg"]["Altitude"][0])
            * self.stat_description["avg"]["Altitude"].unit,
        )
        altitude = altitude_list[
            (altitude_list.to_value() >= floor.to_value(altitude_list.unit))
            & (altitude_list.to_value() < ceiling.to_value(altitude_list.unit))
        ]
        altitude = altitude.to(self.stat_description["avg"]["Altitude"].unit)
        temperature = (
            self._interpolate_cubic(
                self.stat_description["avg"]["Altitude"],
                self.stat_description["avg"]["Temperature"],
                altitude,
            )
            * self.stat_description["avg"]["Temperature"].unit
        )
        relative_humidity = (
            self._interpolate_cubic(
                self.stat_description["avg"]["Altitude"],
                self.stat_description["avg"]["Relative humidity"],
                altitude,
            )
            * self.stat_description["avg"]["Relative humidity"].unit
        )
        pressure = (
            self._interpolate_cubic(
                self.stat_description["avg"]["Altitude"],
                self.stat_description["avg"]["Pressure"],
                altitude,
            )
            * self.stat_description["avg"]["Pressure"].unit
        )
        thickness = pressure / STD_GRAVITATIONAL_ACCELERATION
        density = (
            self._interpolate_cubic(
                self.stat_description["avg"]["Altitude"],
                self.stat_description["avg"]["Density"],
                altitude,
            )
            * self.stat_description["avg"]["Density"].unit
            / N0_AIR
        )
        rel_water_vapor_pressure = (
            partial_pressure_water_vapor(temperature, relative_humidity) / pressure
        ).decompose()
        rel_refractive_index = (
            self._refractive_index(
                pressure, temperature, relative_humidity, 350.0 * u.nm, co2_concentration
            )
            - 1.0
        )

        corsika_input_table = Table()
        tables = []

        for i in np.arange(len(altitude)):
            outdict = {
                "altitude": altitude[i].to(u.km),
                "atmospheric_density": density[i].to(u.g / u.cm**3),
                "atmospheric_thickness": thickness[i].decompose().to(u.g / u.cm**2),
                "refractive_index_m_1": rel_refractive_index[i],
                "temperature": temperature[i],
                "pressure": pressure[i],
                "partial_water_pressure": rel_water_vapor_pressure[i],
            }
            tables.append(outdict)
        # Merge ECMWF profile with upper atmospheric profile
        if reference_atmosphere:
            reference_atmosphere_table = self._pick_up_reference_atmosphere(
                ceiling, floor, reference_atmosphere
            )
            tables.append(reference_atmosphere_table)
        else:
            logger.info(
                "Since reference atmosphere was not provided, the resulting atmospheric model will be constrained to the extent of the provided meteorological data."
            )
        corsika_input_table = vstack(tables)
        corsika_input_table.sort("altitude")
        corsika_input_table.write(outfile, overwrite=True)

    def create_mdp(self, mdp_file):
        """
        Write an output file with the molecular number density per altitude
        """

        altitudes = np.arange(0.0, 20000.0, 1000) * u.m
        altitudes = altitudes.to(self.stat_description["avg"]["Altitude"].unit)
        number_density = (
            self._interpolate_cubic(
                self.stat_description["avg"]["Altitude"],
                self.stat_description["avg"]["Density"],
                altitudes,
            )
            * self.stat_description["avg"]["Density"].unit
        )
        t = Table([altitudes, number_density], names=["altitude", "number density"])
        try:
            t.write(mdp_file, overwrite=True)
        except Exception as ex:
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(ex).__name__, ex.args)
            logger.error(message)
            sys.exit(1)

    def rayleigh_extinction(
        self,
        rayleigh_extinction_file,
        co2_concentration,
        wavelength_min=200 * u.nm,
        wavelength_max=1000 * u.nm,
        reference_atmosphere=None,
        altitude_list=STD_CORSIKA_ALTITUDE_PROFILE,
        rayleigh_scattering_altitude_bins=RAYLEIGH_SCATTERING_ALTITUDE_BINS,
    ):
        """
        Calculates the absolute integral optical depth due to Rayleigh scattering
        per altitude bins as a function of wavelength.
        The optical depth (AOD) for an altitude h over the observatory will be given by
        the integral of the monochromatic volume coefficient beta, with integration limits
        h_obs up to h.

        Parameters:
        -----------
        rayleigh_extinction_file : string
            Name of the returned file with the extinction profile.
        co2_concentration : float
            12MACOBAC value
        wavelength_min : Quantity

        wavelength_max : Quantity

        reference_atmosphere : path
            The path where the file with the reference atmosphere model is located.
        altitude_list : Quantity
            Tuple with the altitudes that the atmospheric parameters will be calculated. Units of length.
        rayleigh_scattering_altitude_bins : Quantity
            Tuple with the altitudes that the AOD will be calculated. Units of length.

        Returns:
        --------
            Ecsv file with absolute optical depth per altitude bin per wavelength bin. The data model is the same with MODTRAN files.
        """

        floor = rayleigh_scattering_altitude_bins[0].to(u.km)
        ceiling = self._take_closest(
            altitude_list.to(self.stat_description["avg"]["Altitude"].unit),
            (self.stat_description["avg"]["Altitude"][0])
            * self.stat_description["avg"]["Altitude"].unit,
        )
        altitude = rayleigh_scattering_altitude_bins.to(u.km)
        if ceiling.to_value(u.km) > 7.0:
            altitude_middle = np.arange(7.0, ceiling.to_value(u.km), 1)
            altitude = (
                np.concatenate((altitude.to_value(u.km), altitude_middle))
            ) * rayleigh_scattering_altitude_bins.unit

        interpolation_centers = (altitude[:-1] + altitude[1:]) / 2
        temperature_lower = (
            self._interpolate_cubic(
                self.stat_description["avg"]["Altitude"].to(u.km),
                self.stat_description["avg"]["Temperature"],
                interpolation_centers.to(u.km),
            )
            * self.stat_description["avg"]["Temperature"].unit
        )
        relative_humidity_lower = (
            self._interpolate_cubic(
                self.stat_description["avg"]["Altitude"].to(u.km),
                self.stat_description["avg"]["Relative humidity"],
                interpolation_centers.to(u.km),
            )
            * self.stat_description["avg"]["Relative humidity"].unit
        )
        pressure_lower = (
            self._interpolate_cubic(
                self.stat_description["avg"]["Altitude"].to(u.km),
                self.stat_description["avg"]["Pressure"],
                interpolation_centers.to(u.km),
            )
            * self.stat_description["avg"]["Pressure"].unit
        )

        # Concatenate with reference atmosphere
        if reference_atmosphere:
            reference_atmosphere_table = self._pick_up_reference_atmosphere(
                ceiling, floor, reference_atmosphere
            )
            length_of_columns = len(reference_atmosphere_table)
            relative_humidity_upper = (
                np.zeros(length_of_columns) * self.stat_description["avg"]["Relative humidity"].unit
            )
            try:
                relative_humidity = np.concatenate(
                    (relative_humidity_lower, relative_humidity_upper)
                )
                pressure = np.concatenate((pressure_lower, reference_atmosphere_table["pressure"]))
                temperature = np.concatenate(
                    (temperature_lower, reference_atmosphere_table["temperature"])
                )
                altitude = np.concatenate((altitude, reference_atmosphere_table["altitude"]))
            except Exception as ex:
                template = "An exception of type {0} occurred. Arguments:\n{1!r}"
                message = template.format(type(ex).__name__, ex.args)
                logger.error(message)
        else:
            logger.info(
                "Since the reference atmosphere was not provided, the resulting atmospheric model will be constrained to the extent of the provided meteorological data."
            )
            temperature = temperature_lower
            pressure = pressure_lower
            relative_humidity = relative_humidity_lower
        t = QTable(
            [altitude[1:], pressure, temperature, relative_humidity],
            names=("altitude", "pressure", "temperature", "relative_humidity"),
        )
        t.sort("altitude")
        bin_widths = np.diff(np.sort(altitude))
        t["bin_widths"] = bin_widths
        mask = t["altitude"] > floor
        wavelength_range = (
            np.arange(wavelength_min.to_value(u.nm), wavelength_max.to_value(u.nm), 1) * u.nm
        )
        rayleigh_extinction_table = Table(names=wavelength_range)
        col_alt_max = Column(name="altitude_max")
        col_alt_min = Column(name="altitude_min")
        rayleigh_extinction_table.add_columns([col_alt_max, col_alt_min], indexes=[0, 0])
        aod_dict = {
            aod: 0
            for aod in np.arange(wavelength_min.to_value(u.nm), wavelength_max.to_value(u.nm))
        }
        for row in t[mask]:
            new_row = []
            new_row.append(row["altitude"])
            new_row.append(row["altitude"] - row["bin_widths"])
            for wavelength in wavelength_range:
                rayleigh = Rayleigh(
                    wavelength,
                    co2_concentration,
                    row["pressure"],
                    row["temperature"],
                    row["relative_humidity"],
                )
                beta = rayleigh.beta
                aod = row["bin_widths"] * beta
                aod_dict[wavelength.to_value(u.nm)] += aod
                new_row.append(aod_dict[wavelength.to_value(u.nm)])
            rayleigh_extinction_table.add_row(new_row)
        rayleigh_extinction_table.write(rayleigh_extinction_file, overwrite=True)
        return rayleigh_extinction_file



"""
Set of utilities for decoding and analysing grib files.
"""

import glob
import os
import logging
import logging.config
import numpy as np
import pygrib as pg
import astropy.units as u
from astropy.coordinates import Angle, Latitude, Longitude
from astropy.time import Time
from astropy.table import Table, join, vstack
from molecularprofiles.utils.constants import (
    STD_NUMBER_DENSITY,
    STD_AIR_PRESSURE,
    STD_AIR_TEMPERATURE,
    DENSITY_SCALE_HEIGHT,
    STD_GRAVITATIONAL_ACCELERATION,
    STD_EARTH_RADIUS,
)

ROOTDIR = os.path.dirname(os.path.abspath(__file__))
log_config_file = f"{ROOTDIR}/mdps_log.conf"
logging.config.fileConfig(fname=log_config_file, disable_existing_loggers=False)
logger = logging.getLogger(__name__)


def get_altitude_from_geopotential_height(geopotential_height, latitude):
    """
    Function to compute the real altitude from the geopotential value at
    a certain coordinate on Earth.
    Uses expression 20 from http://dx.doi.org/10.1029/2002JB002333

    Parameters
    ----------
    geopotential : astropy.units.Quantity [u.m]
        Geopotential height
    latitude : astropy.coordinates.Latitude
        Geographical latitude of interest.

    Returns
    -------
    astropy.units.Quantity [u.m]
        Real altitude as fGeoidOffset.
    """

    lat = (np.asarray(latitude) * latitude.unit).to_value(u.rad)
    return (
        (1.0 + 0.002644 * np.cos(2 * lat)) * (geopotential_height)
        + (1 + 0.0089 * np.cos(2 * lat)) * ((geopotential_height) ** 2 / STD_EARTH_RADIUS.to(u.m))
    ).to(u.m)


def get_gribfile_variables(filename):
    """
    Returns all the different variable names in a grib file.

    Parameters
    ----------
    filename : str
       Path to the grib file.

    Returns
    -------
    list
        varname (str): variable names.
    list
        varshortname (str): variable short names.
    """
    logger.debug("Opening %s grib file", filename)
    with pg.open(filename) as grib_file:
        varshortname = []
        varname = []
        grib_file.read(10)[0]
        while True:
            v = grib_file.read(1)[0]
            variable_name = v.name.replace(" ", "")
            variable_short_name = v.shortName
            if variable_name not in varname:
                varname.append(variable_name)
                varshortname.append(variable_short_name)
            else:
                break
    return varname, varshortname


def get_plevels(atmo_parameter):
    """
    The measurements of atmospheres parameters are taking place in various pressure levels.
    This functions returns the indices of these levels where we have measurements
    for the given atmospheric parameter.
    E.g. for GDAS data that should be a range of values [1, 23].

    Parameters
    ----------
    atmo_parameter : str FIXME: this description doesn't
        Atmospheric parameter from grib file.

    Returns
    -------
    list
        List (int) with the indices of the pressure levels
    """

    plevels = []
    index = []
    for i in range(len(atmo_parameter)):
        pressure_level = atmo_parameter[i].level
        if pressure_level not in plevels:
            plevels.append(pressure_level)
            index.append(i + 1)
        else:
            break
    return index


def create_table(grib_var):
    """
    Create astropy.table.Table from grib record

    Parameters
    ----------
    grib_var :
        Grib data record

     Returns
     -------
     astropy.table.Table
         Table with the different measurables together with their dimensions
    """
    res = Table()
    for v in grib_var:
        unit = u.Unit(v.units)
        timestamp = Time(
            {"year": v.year, "month": v.month, "day": v.day, "hour": v.hour}, scale="utc"
        )
        pressure_level = v.level * u.hPa
        latitudes = Latitude(v.latlons()[0].ravel() * u.deg)
        longitudes = Longitude(v.latlons()[1].ravel() * u.deg, wrap_angle=180 * u.deg)
        if isinstance(v.values, float):
            vals = np.array([v.values]) * unit
        else:
            vals = v.values.ravel() * unit
        t = Table([latitudes, longitudes], names=["Latitude", "Longitude"])
        t["Timestamp"] = timestamp
        t["Pressure"] = pressure_level
        t[v.name] = vals
        res = vstack([res, t])
    return res


def get_grib_file_data(filename):
    """
    This function opens a grib file, selects the parameters
    (all available: Temperature, Geopotential, RH, ...),
    and creates an astropy.table.Table with them.
    Parameters
    ----------
    filename : str
        Path to the grib file.

    Returns
    -------
    astropy.table.Table
        Table with grib file data gridded in isobaric levels
    """

    logger.debug("Working with %s", filename)
    logger.debug("Getting all variable names...")
    variable_names, variable_short_names = get_gribfile_variables(filename)
    logger.info("Indexing the file %s (this might take a while...)", filename)
    grib_file = pg.index(filename, "shortName", "typeOfLevel")
    logger.debug(
        "Selecting the parameters information for %s (this might take a while...)", filename
    )
    gpm = u.def_unit("gpm", u.m)
    u.add_enabled_units([gpm])
    data = Table()
    for short_name in variable_short_names:
        if short_name == "unknown":
            continue
        var = grib_file.select(shortName=short_name, typeOfLevel="isobaricInhPa")
        try:
            t = create_table(var)
        except ValueError:
            logger.warning("Grib message can't be parsed, skipping")
            continue
        if len(data) == 0:
            data = t
        else:
            data = join(data, t, join_type="outer")
    return data


def extend_grib_data(data):
    """
    Extends grib data table by filling the gaps in data and calculating additional quantities:
        - altitude
        - density
        - exponential density
        - wind direction

    Parameters
    ----------
    astropy.table.Table
        Table with grib data
    Returns
    -------
    astropy.table.Table
        Extended table with grib data and additional quantities
    """
    logger.debug("Check for gaps in relative humidity and fill them if neccessary")
    data["Relative humidity"] = data["Relative humidity"].filled(0)
    logger.debug("Compute altitude from geopotential")
    if "Geopotential Height" in data.keys():
        data["Altitude"] = get_altitude_from_geopotential_height(
            data["Geopotential Height"], data["Latitude"]
        )
    else:
        data["Altitude"] = get_altitude_from_geopotential_height(
            data["Geopotential"].quantity / STD_GRAVITATIONAL_ACCELERATION, data["Latitude"]
        )
    logger.debug("Compute density")
    data["Density"] = (
        STD_NUMBER_DENSITY
        * data["Pressure"]
        / STD_AIR_PRESSURE
        * STD_AIR_TEMPERATURE
        / data["Temperature"]
    )
    logger.debug("Compute exponential density")
    data["Exponential Density"] = (
        data["Density"] / STD_NUMBER_DENSITY * np.exp(data["Altitude"] / DENSITY_SCALE_HEIGHT)
    )
    logger.debug("Compute wind speed")
    data["Wind Speed"] = np.sqrt(
        data["U component of wind"] ** 2 + data["V component of wind"] ** 2
    )
    logger.debug("Compute wind direction")
    data["Wind Direction"] = Angle(
        np.array(
            np.arctan2(-data["V component of wind"].value, -data["U component of wind"].value)
        ),
        unit=u.rad,
    ).wrap_at(360 * u.deg)
    return data


def save_grib_table(data, filename, fmt="ecsv"):
    """
    Save grib data in a file according to provided format.

    Parameters
    ----------
    data: astropy.table.Table
        Grib data in astropy table
    filename: str
        Path to the file
    fmt: str
        Desired format
    """
    if fmt == "ecsv":
        data.write(filename, format="ascii.ecsv", overwrite=True)
    elif fmt == "magic":
        raise NotImplementedError("Magic txt format writing is not implemented yet")
    else:
        raise ValueError("Not recognized format")


def convert_to_text(data_path):
    """
    Converts GRIB/GRIB2 files to ecsv

    Parameters
    ----------
    data_path: path-like
        Path to the folder with GRIB/GRIB2 data files
    """

    grib_file_list = list(glob.glob(os.path.join(data_path, "*.grib2"))) + list(
        glob.glob(os.path.join(data_path, "*.grib"))
    )
    for grib_file in grib_file_list:
        data = get_grib_file_data(grib_file)
        data = extend_grib_data(data)
        save_grib_table(data, os.path.splitext(grib_file)[0] + ".ecsv")


def merge_ecsv_files(data_path):
    """
    Merge ecsv files

    Parameters
    ----------
    data_path: path-like
        Path to the folder with GRIB/GRIB2 data files
    """
    merged_data = Table()
    for tfile in os.listdir(data_path):
        if tfile.endswith(".ecsv"):
            data = Table.read(f"{data_path}/{tfile}")
            merged_data = vstack([merged_data, data])
    save_grib_table(merged_data, f"{data_path}/merged_file.ecsv")



import functools
import numpy as np
import astropy.units as u
from molecularprofiles.utils import humidity
from molecularprofiles.utils.constants import (
    STD_NUMBER_DENSITY,
    STD_AIR_PRESSURE,
    STD_AIR_TEMPERATURE,
    STD_RELATIVE_HUMIDITY,
    NITROGEN_RATIO,
    OXYGEN_RATIO,
    ARGON_RATIO,
    GAS_CONSTANT,
)


class Rayleigh:
    """
    A small Rayleigh-scattering program based on:
    C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari
    "Improved algorithm for calculations of Rayleigh-scattering optical
    depth in standard atmospheres"
    Applied Optics 44 Nr. 16 (2005) 3320

    The calculation of refractive index is based on:
    P.E. Ciddor, "Refractive index of air: new equations for hte visible and near infrared",
    Applied Optics 35 (1996) 1566

    P.E. Ciddor, "Refractive index of air: 3. The roles of CO2, H20 and refractivity virals",
    Applied Optics 41 (2002) 2292

    The principal King factor formula is based on:
    D.R. Bates, "Rayleigh scattering by air",
    Planet. Space Sci. 32 (1984) 785

    B.A. Bodhaine, N.B. Wood, E.G. Dutton, J.R. Slusser, "On Rayleigh optical depth calculations",
    J. Atmosph. Osceanic Technol. 16 (1999) 1854

    The calculation of the Chandrasekhar phase function is based on:
    S. Chandrasekhar, Radiative Transfer, Dover Publications, 1960.

    E.J. McCartney, "Optics of the Atmosphere. Scattering by Molecules and Particles"
    Wiley & Sons, New York, 1977.

    Adapted from MRayleigh, written by Markus Gaug <markus.gaug@uab.cat>, 04/2013
    .. moduleauthor:: Scott Griffiths <sgriffiths@ifae.es>
    """

    def __init__(
        self,
        wavelength,
        co2_bkg,
        pressure=STD_AIR_PRESSURE,
        temperature=STD_AIR_TEMPERATURE,
        relative_humidity=STD_RELATIVE_HUMIDITY,
    ):
        """
        Constructor for Rayleigh-scattering class.

        Parameters
        ----------
        wavelength : astropy.units.Quantity
            Wavelength of light.
        co2_bkg : float
            CO2 concentration [ppmv]
        pressure : astropy.units.Quantity
            Atmospheric pressure
        temperature : astropy.units.Quantity
            Air temperature
        relative_humidity : astropy.units.Quantity
            Relative humidity [%]
        """
        # check inputs for bad values
        if np.any(wavelength < 200 * u.nm) or np.any(wavelength > 4000 * u.nm):
            raise ValueError("Wavelength range only from 200 nm - 4 micrometer allowed.")
        if np.any(pressure < 0 * u.hPa) or np.any(pressure > 1400 * u.hPa):
            raise ValueError("Pressure only in range 0 - 1400 hPa allowed.")
        if np.any(temperature < 150 * u.K) or np.any(temperature > 400 * u.K):
            raise ValueError("Temperatures only in range 150 - 400 K allowed.")
        # if np.any(relative_humidity < 0) or np.any(relative_humidity > 100 * u.percent):
        #    raise ValueError("Relative humity must lie between 0 - 100.")
        # FIXME: Relative humidity can exceed 100%, in case of supersaturated air.
        # The maximum limit should be thought better. However relative humidity cannot go negative.
        # If we receive negative values it is most probably due to interpolation problems.
        # Till we fix these issues it is better to comment out the minimum limit.
        if co2_bkg < 200 or co2_bkg > 1000:
            raise ValueError("CO2 concentrations only in range 200 - 1000 ppmv allowed.")

        self.wavelength = wavelength.to(u.nm)
        self.pressure = pressure.to(u.hPa)
        self.temperature = temperature.to(u.K, equivalencies=u.temperature())
        self.relative_humidity = relative_humidity.to(u.percent)
        self.co2_bkg = co2_bkg  # [ppmv]  CO2 concentration of air

    @functools.cached_property
    def molecular_number_density(self):
        """
        Returns
        -------
        astropy.units.Quantity
            Molecular number density [cm^-3]

        Notes
        -----
        See Eq. 3 in [1]_.

        References
        ----------
        .. [1] C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari
           "Improved algorithm for calculations of Rayleigh-scattering optical depth
           in standard atmospheres", Applied Optics 44 Nr. 16 (2005) 3320
        """
        return (
            (
                STD_NUMBER_DENSITY
                * self.pressure
                / STD_AIR_PRESSURE
                * STD_AIR_TEMPERATURE
                / self.temperature
            )
            .decompose()
            .to(1 / u.cm**3)
        )

    @functools.cached_property
    def scattering_cross_section(self):
        """
        Returns
        -------
        astropy.units.Quantity
            Total Rayleigh scattering cross section per molecule [cm^2].

        Notes
        -----
        See Eq. 4 in [1]_.

        References
        ----------
        .. [1] C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari
           "Improved algorithm for calculations of Rayleigh-scattering optical depth
           in standard atmospheres", Applied Optics 44 Nr. 16 (2005) 3320
        """
        return (
            (
                24
                * np.pi**3
                * (self.refractive_index**2 - 1) ** 2
                / (
                    self.wavelength**4
                    * self.molecular_number_density**2
                    * (self.refractive_index**2 + 2) ** 2
                )
                * self.king_factor
            )
            .decompose()
            .to(u.cm**2)
        )

    @functools.cached_property
    def beta(self):
        """
        Calculates the monochromatic volume coefficient for the total molecular
        scattering in cloudless air, beta.

        Returns
        -------
        astropy.units.Quantity
            Monochromatic volume coefficient for the total molecular scattering
            in cloudless air [1/km]

        Notes
        -----
        See Eq. 2 in [1]_.

        References
        ----------
        .. [1] C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari
           "Improved algorithm for calculations of Rayleigh-scattering optical depth
           in standard atmospheres", Applied Optics 44 Nr. 16 (2005) 3320
        """
        return (
            (self.molecular_number_density * self.scattering_cross_section).decompose().to(1 / u.km)
        )

    @functools.cached_property
    def refractive_index(self):
        """
        Ciddor formula for calculation of refractive index in moist air.
        The obtained refractive index is precise to 1e-7.

        Returns
        -------
        float
            Index of refraction of moist air

        Notes
        -----
        Cross-checked with:
        http://emtoolbox.nist.gov/Wavelength/Documentation.asp#IndexofRefractionofAir
        """
        refractive_index_dry = (
            1e-8
            * (
                5792105 / (238.0185 - 1 / self.wavelength.to_value(u.micron) ** 2)
                + 167917 / (57.362 - 1 / self.wavelength.to_value(u.micron) ** 2)
            )
            + 1
        )  # Tomasi eq. 17

        # refractive index of dry air at standard p and T, for given C (e = 0)
        refractive_index_dry_std_air = (1 + 0.534e-6 * (self.co2_bkg - 450)) * (
            refractive_index_dry - 1
        ) + 1  # Tomasi eq. 18

        # refractive index of pure water vapor at standard T and e
        # (T* = 293.15 K = 20 C, and e* = 1333 Pa)
        refractive_index_water_vapour = (
            1.022e-8
            * (
                295.235
                + 2.6422 / self.wavelength.to_value(u.micron) ** 2
                - 0.032380 / self.wavelength.to_value(u.micron) ** 4
                + 0.004028 / self.wavelength.to_value(u.micron) ** 6
            )
            + 1
        )  # Tomasi eq. 19

        # calculate the respective densities (see Tomasi et al., pp. 3325 ff)
        molar_mass_dry_air = (
            1e-3 * (28.9635 + 12.011e-6 * (self.co2_bkg - 400)) * u.kg / u.mol
        )  # Tomasi eq. 13
        molar_mass_water_vapour = 0.018015 * u.kg / u.mol
        molar_fraction_water_vapour = humidity.molar_fraction_water_vapor(
            self.pressure, self.temperature, self.relative_humidity
        )  # molar fraction of water vapor in moist air
        compressibility_dry_air = humidity.compressibility(
            STD_AIR_PRESSURE, STD_AIR_TEMPERATURE, 0
        )  # compressibility of dry air
        compressibility_water_vapour = humidity.compressibility(
            1333 * u.Pa, 293.15 * u.K, 1
        )  # compressibility of pure water vapor
        compressibility_moist_air = humidity.compressibility(
            self.pressure, self.temperature, molar_fraction_water_vapour
        )  # compressibility of moist air

        # density of dry air at standard p and T
        dry_air_density_stdpt = humidity.density_moist_air(
            STD_AIR_PRESSURE, STD_AIR_TEMPERATURE, compressibility_dry_air, 0, self.co2_bkg
        )

        # density of pure water vapor at at standard T and e (T* = 293.15 K = 20 C, and e* = 1333 Pa)
        water_vapour_density_stdpt = humidity.density_moist_air(
            1333 * u.Pa, 293.15 * u.K, compressibility_water_vapour, 1, self.co2_bkg
        )

        # density of the dry component of moist air
        density_dry_comp_moist_air = (
            self.pressure
            * molar_mass_dry_air
            * (1 - molar_fraction_water_vapour)
            / (compressibility_moist_air * GAS_CONSTANT * self.temperature)
        ).decompose()

        # density of the water vapor component of moist air
        density_water_vapour_moist_air = (
            self.pressure
            * molar_mass_water_vapour
            * molar_fraction_water_vapour
            / (compressibility_moist_air * GAS_CONSTANT * self.temperature)
        ).decompose()

        return (
            1
            + (density_dry_comp_moist_air / dry_air_density_stdpt)
            * (refractive_index_dry_std_air - 1)
            + (density_water_vapour_moist_air / water_vapour_density_stdpt)
            * (refractive_index_water_vapour - 1)
        )  # Ciddor eq. 5, Tomasi eq. 11

    @functools.cached_property
    def king_factor(self):
        """
        Calculates the current best estimate of the King factor of moist air.

        Returns
        -------
        float
            King factor [dimensionless]

        Notes
        -----
        The King factor is used to take into account effects due to the anisotropic
        properties of air molecules since anisotropic molecules scatter more radiation
        at 90 degrees scattering angles than isotropic molecules with the same index
        of refraction.

        Precision not stated in Tomasi et al., but probably better than 1e-4.
        Effects of relative_humidity are of the order of several times 1e-4.
        """
        water_vapour_partial_pressure = humidity.partial_pressure_water_vapor(
            self.temperature, self.relative_humidity
        )  # water vapor partial pressure [hPa]

        king_factor_n2 = (
            1.034 + 3.17e-4 / self.wavelength.to_value(u.micron) ** 2
        )  # partial King factor for N2 molecules
        king_factor_o2 = (
            1.096
            + 1.385e-3 / self.wavelength.to_value(u.micron) ** 2
            + 1.448e-4 / self.wavelength.to_value(u.micron) ** 4
        )  # partial King factor for O2 molecules
        king_factor_ar = 1.00  # partial King factor for Ar molecules
        king_factor_co2 = 1.15  # partial King factor for CO2 molecules
        king_factor_wv = 1.001  # partial King factor for water vapor

        co2_ratio = 1e-6 * self.co2_bkg  # CO2
        water_vapour_ratio = (water_vapour_partial_pressure / self.pressure).decompose().to_value()

        return (
            NITROGEN_RATIO * king_factor_n2
            + OXYGEN_RATIO * king_factor_o2
            + ARGON_RATIO * king_factor_ar
            + co2_ratio * king_factor_co2
            + water_vapour_ratio * king_factor_wv
        ) / (
            NITROGEN_RATIO + OXYGEN_RATIO + ARGON_RATIO + co2_ratio + water_vapour_ratio
        )  # Tomasi eq. 22

    @functools.cached_property
    def depolarization(self):
        """
        Current best estimate of the depolarization factor of moist air.

        Precision not stated in Tomasi et al., but probably better than 1e-4.
        Effects of relative_humidity are of the order of several times 1e-4.

        Returns
        -------
        float
            Depolarization factor of moist air.
        """
        return (
            6 * (self.king_factor - 1) / (3 + 7 * self.king_factor)
        )  # Tomasi eq. 5, solved for rho

    def phase_function(self, angle):
        """
        Calculates the Chandrasekhar phase function according to Eq. 255 of [1]_.

        Parameters
        ----------
        angle : astropy.units.Quantity
            Scattering angle.

        Returns
        -------
            Chandrasekhar phase function for scattering of natural light.

        References
        ----------
        .. [1] S. Chandrasekhar, Radiative Transfer, Dover Publications, 1960.
        """
        rho = self.depolarization

        # need to solve Chandrasekhar eq. 254 for gamma as a function of rho
        f_1 = (2 + 2 * rho) / (2 + rho)
        f_2 = (1 - rho) / (1 + rho)
        return 0.75 * f_1 * (1 + f_2 * np.cos(angle) ** 2)  # Chandrasekhar eq. 255

    # TODO: where does this come from? Needs unit test
    def back_scattering_coefficient(self, angle):
        """
        Back-scattering coefficient for a given scattering angle.

        Parameters
        ----------
        angle : astropy.units.Quantity
            Scattering angle.

        Returns
        -------
        astropy.units.Quantity
            Back-scattering coefficient [1/km]
        """
        return self.phase_function(angle) * self.beta / (4 * np.pi)

    def print_params(self):
        """Prints Rayleigh scattering parameters."""
        print(f"Wavelength:              {self.wavelength}")
        print(f"Air Pressure:            {self.pressure}")
        print(f"Air Temperature:         {self.temperature}")
        print(f"Rel. Humidity:           {self.relative_humidity}")
        print(f"CO2 concentration:       {self.co2_bkg} ppmv")
        print(f"Refractive Index:        {self.refractive_index}")
        print(f"King Factor:             {self.king_factor}")
        print(f"Depolarization:          {self.depolarization}")
        print(f"Mol. cross section:      {self.scattering_cross_section}")
        print(f"Volume scattering coeff: {self.beta}")



import astropy.units as u
from astropy.constants import N_A

# standard atmosphere thermodynamic values
STD_NUMBER_DENSITY = (
    2.546899e19 / u.cm**3
)  # [cm^-3] molecular number density for standard air conditions
STD_AIR_PRESSURE = 1013.25 * u.hPa  # [hPa]   standard air pressure
STD_AIR_TEMPERATURE = 288.15 * u.K  # [K]     standard air temperature
STD_AIR_DENSITY = 1.225 * u.kg / u.m**3  # [kg/m^3] standard air density
DENSITY_SCALE_HEIGHT = 9500.0 * u.km  # [km]    density scale hight for La Palma Winter
STD_RELATIVE_HUMIDITY = 45.9 * u.percent  # [%]     standard air rel. humidity
# atmospheric composition
NITROGEN_RATIO = 0.78084 * u.dimensionless_unscaled
OXYGEN_RATIO = 0.20946 * u.dimensionless_unscaled
ARGON_RATIO = 0.00934 * u.dimensionless_unscaled
GAS_CONSTANT = 8.31451 * u.J / (u.K * u.mol)  # gas constant [J/mol/K]
MOLAR_MASS_WATER_VAPOR = 0.018015 * u.kg / u.mol  # molar mass of water vapor [kg/mol]
MOLAR_MASS_AIR = 0.0289644 * u.kg / u.mol  # molar mass of air [kg/mol]
# misc physics constants
STD_GRAVITATIONAL_ACCELERATION = (
    9.80665 * u.m / u.s**2
)  # standard acceleration of free fall [m/s^2]
STD_EARTH_RADIUS = 6245 * u.km  # Earth radius for geopotential height conversion [km]
N0_AIR = N_A / MOLAR_MASS_AIR
STD_CORSIKA_ALTITUDE_PROFILE = (
    0.0,
    1000.0,
    2000.0,
    3000.0,
    4000.0,
    5000.0,
    6000.0,
    7000.0,
    8000.0,
    9000.0,
    10000.0,
    11000.0,
    12000.0,
    13000.0,
    14000.0,
    15000.0,
    16000.0,
    17000.0,
    18000.0,
    19000.0,
    20000.0,
    21000.0,
    22000.0,
    23000.0,
    24000.0,
    25000.0,
    26000.0,
    27000.0,
    28000.0,
    29000.0,
    30000.0,
    32000.0,
    34000.0,
    36000.0,
    38000.0,
    40000.0,
    42000.0,
    44000.0,
    46000.0,
    48000.0,
    50000.0,
    55000.0,
    60000.0,
    65000.0,
    70000.0,
    75000.0,
    80000.0,
    85000.0,
    90000.0,
    95000.0,
    100000.0,
    105000.0,
    110000.0,
    115000.0,
    120000.0,
) * u.m
RAYLEIGH_SCATTERING_ALTITUDE_BINS = (
    2.1,
    2.2,
    2.3,
    2.4,
    2.6,
    2.8,
    3.1,
    3.5,
    4.0,
    4.5,
    5.0,
    5.5,
    6.0,
) * u.km



"""
Calculates the properties of moist air, including:
* density
* compressibility
* enhancement factor
* saturation vapor pressure
* molar fraction from relative humidity
* partial pressure from relative humidity

Adapted from MHumidity, written by Markus Gaug <markus.gaug@uab.cat>, 04/2013
"""

import astropy.units as u
import numpy as np
from molecularprofiles.utils.constants import GAS_CONSTANT, MOLAR_MASS_WATER_VAPOR


def compressibility(pressure, temperature, x_w):
    """
    Calculates the compressibility of moist air.

    Parameters
    ----------
    pressure : astropy.units.Quantity
        Atmospheric pressure
    temperature : astropy.units.Quantity
        Air temperature
    x_w : float
        Molar fraction of water vapor

    Returns
    -------
    float
        compressibility of moist air (dimensionless constant, 0 < Z < 1)

    Notes
    -----
    Implemented according to Eq. 5 of [1]_. Notation and constants according to
    Eq. 16 of [2]_.

    References
    ----------
    .. [1] R.S. Davis, "Equation for the determination of the density of moist air" Metrologia, 29 (1992) 67-70
    .. [2] C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari "Improved algorithm
       for calculations of Rayleigh-scattering optical depth in standard atmospheres",
       Applied Optics 44 Nr. 16 (2005) 3320
    """
    p = pressure.to(u.Pa)
    T = temperature.to(u.K, equivalencies=u.temperature())
    a_0 = 1.58123e-6 * u.K / u.Pa
    a_1 = -2.9331e-8 / u.Pa
    a_2 = 1.1043e-10 / (u.K * u.Pa)
    b_0 = 5.707e-6 * u.K / u.Pa
    b_1 = -2.051e-8 / u.Pa
    c_0 = 1.9898e-4 * u.K / u.Pa
    c_1 = -2.376e-6 / u.Pa
    d_0 = 1.83e-11 * u.K**2 / u.Pa**2
    d_1 = -7.65e-9 * u.K**2 / u.Pa**2

    return (
        1
        - p
        / T
        * (
            a_0
            + a_1 * (T - 273.15 * u.K)
            + a_2 * (T - 273.15 * u.K) ** 2
            + (b_0 + b_1 * (T - 273.15 * u.K)) * x_w
            + (c_0 + c_1 * (T - 273.15 * u.K)) * x_w**2
        )
        + p**2 / T**2 * (d_0 + d_1 * x_w**2)
    ).to_value()


def enhancement_factor(pressure, temperature):
    """
    Calculates the enhancement factor of water vapor in air.

    Parameters
    ----------
    pressure : astropy.units.Quantity
        Atmospheric pressure
    temperature : astropy.units.Quantity
        Air temperature

    Returns
    -------
    float
        Enhancement factor (dimensionless constant)

    Notes
    -----
    Calculated according to Eq. 14 of [1]_.

    References
    ----------
    .. [1] C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari "Improved algorithm
       for calculations of Rayleigh-scattering optical depth in standard atmospheres",
       Applied Optics 44 Nr. 16 (2005) 3320
    """
    p = pressure.to(u.Pa)
    T = temperature.to(u.K, equivalencies=u.temperature())
    return (1.00062 + 3.14e-8 * p / u.Pa + 5.6e-7 * (T - 273.15 * u.K) ** 2 / u.K**2).to_value()


# TODO: This function needs a simple unit test to check which function is called
# TODO: Goff-Gratch doesn't seem to be as good as Buck, or maybe Wexler
def saturation_vapor_pressure(temperature):
    """
    Calculates the vapor pressure at saturation.

    Parameters
    ----------
    temperature : astropy.units.Quantity
        Air temperature

    Returns
    -------
    astropy.units.Quantity
        Saturation vapor pressure

    Notes
    -----
    In case temperatures above 0 deg C it follows [1]_, see also Eq. 15 in [2]_.
    Otherwise, Goff-Gratch equation (Eq. 1 in [3]_) is used.

    References
    ----------
    .. [1] R.S. Davis, "Equation for the determination of the density of moist air" Metrologia, 29 (1992) 67-70
    .. [2] C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari "Improved algorithm
       for calculations of Rayleigh-scattering optical depth in standard atmospheres",
       Applied Optics 44 Nr. 16 (2005) 3320
    .. [3] http://cires.colorado.edu/~voemel/vp.html
    """
    T = temperature.to(u.K, equivalencies=u.temperature())
    mask = T > 273.15 * u.K
    res = np.empty(T.shape)
    res[mask] = (
        np.exp(
            1.2378847e-5 / u.K**2 * T[mask] ** 2
            - 1.9121316e-2 / u.K * T[mask]
            + 33.93711047
            - 6343.1645 * u.K / T[mask]
        )
        / 100.0
        # * u.Pa
    )

    theta = (373.16 * u.K / T[~mask]).to_value()  # ratio of steam point (100 deg C) to temperature
    res[~mask] = (
        np.power(
            10,
            (
                -7.90298 * (theta - 1)
                + 5.02808 * np.log10(theta)
                - 1.3816e-7 * (np.power(10, 11.344 * (1 - 1 / theta)) - 1)
                + 8.1328e-3 * (np.power(10, -3.49149 * (theta - 1)) - 1)
                + np.log10(1013.246)
            ),
        )
        # * u.hPa
    )

    return res * u.hPa


# TODO: Check this function and write unit test
def saturation_vapor_pressure_over_water(temperature):
    """
    Calculates the vapor pressure at saturation over water.

    Parameters
    ----------
    temperature : astropy.units.Quantity
        Air temperature

    Returns
    -------
    astropy.units.Quantity
        Saturation vapor pressure

    Notes
    -----
    Implemented according to [1]_. See also [2]_ and [3]_.

    References
    ----------
    .. [1] International Association for the Properties of Water and Steam,
       Peter H. Huang, "New equations for water vapor pressure in the temperature
       range -100 deg. C to 100 deg. C for use with the 1997 NIST/ASME steam tables"
       Papers and abstracts from the third international symposium on humidity and
       moisture, Vol. 1, p. 69-76, National Physical Laboratory, Teddington,
       Middlesex, UK, April 1998.
    .. [2] http://cires.colorado.edu/~voemel/vp.html
    .. [3] https://emtoolbox.nist.gov/wavelength/documentation.asp#AppendixA
    """
    T = temperature.to(u.K, equivalencies=u.temperature()).to_value()
    k_1 = 1.16705214528e03
    k_2 = -7.24213167032e05
    k_3 = -1.70738469401e01
    k_4 = 1.20208247025e04
    k_5 = -3.23255503223e06
    k_6 = 1.49151086135e01
    k_7 = -4.82326573616e03
    k_8 = 4.05113405421e05
    k_9 = -2.38555575678e-01
    k_10 = 6.50175348448e02

    omega = T + k_9 / (T - k_10)
    a = omega**2 + k_1 * omega + k_2
    b = k_3 * omega**2 + k_4 * omega + k_5
    c = k_6 * omega**2 + k_7 * omega + k_8
    x = -b + np.sqrt(b**2 - 4 * a * c)
    return 1e6 * np.power(2 * c / x, 4) * u.Pa


# TODO: Check this function and write unit test
def saturation_vapor_pressure_over_ice(temperature):
    """
    Calculates the vapor pressure at saturation over ice.

    Parameters
    ----------
    temperature: astropy.units.Quantity
        Air temperature

    Returns
    -------
    astropy.units.Quantity
        Saturation vapor pressure

    Notes
    -----
    Implemented according to [1]_. See also [2]_ and [3]_.

    References
    ----------
    .. [1] International Association for the Properties of Water and Steam,
       Peter H. Huang, "New equations for water vapor pressure in the temperature
       range -100 deg. C to 100 deg. C for use with the 1997 NIST/ASME steam tables"
       Papers and abstracts from the third international symposium on humidity and
       moisture, Vol. 1, p. 69-76, National Physical Laboratory, Teddington,
       Middlesex, UK, April 1998.
    .. [2] http://cires.colorado.edu/~voemel/vp.html
    .. [3] https://emtoolbox.nist.gov/wavelength/documentation.asp#AppendixA
    """
    theta = temperature.to(u.K, equivalencies=u.temperature()) / 273.16 * u.K
    a_1 = -13.928169
    a_2 = 34.7078238
    y = a_1 * (1 - np.power(theta, -1.5)) + a_2 * (1 - np.power(theta, -1.25))
    return 611.657 * np.exp(y) * u.Pa


def molar_fraction_water_vapor(pressure, temperature, relative_humidity):
    """
    Calculates the molar fraction of water vapor in moist air.

    Parameters
    ----------
    pressure : astropy.units.Quantity
        Atmospheric pressure
    temperature : astropy.units.Quantity
        Air temperature
    relative_humidity : float
        Relative humidity in percent

    Returns
    -------
    float
        Molar fraction of water vapor in moist air (dimensionless)

    Notes
    -----
    See the text above Eq. 14 in [1]_.

    References
    ----------
    .. [1] C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari
       "Improved algorithm for calculations of Rayleigh-scattering optical depth
       in standard atmospheres", Applied Optics 44 Nr. 16 (2005) 3320
    """
    factor = enhancement_factor(pressure, temperature)
    psv = saturation_vapor_pressure(temperature)
    return (
        factor * relative_humidity.to(u.dimensionless_unscaled) * (psv / pressure).decompose()
    ).to_value()


def density_moist_air(pressure, temperature, moist_air_compressibility, x_w, co2_bkg):
    """
    Density equation of moist air, according to Eq. 1 of [1]_.

    Parameters
    ----------
    pressure : astropy.units.Quantity
        Atmospheric pressure
    temperature : astropy.units.Quantity
        Temperature
    moist_air_compressibility : float
        Compressibility (see compressibility() in this module)
    x_w : float
        Molar fraction of water vapor
    co2_bkg : float
        CO2 volume concentration in ppmv (different unit than in Davis!)

    Returns
    -------
    astropy.units.Quantity
        Density of moist air

    References
    ----------
    .. [1] R.S. Davis, "Equation for the determination of the density of moist air"
       Metrologia, 29 (1992) 67-70
    """
    R = GAS_CONSTANT
    m_w = MOLAR_MASS_WATER_VAPOR
    m_a = (
        1e-3 * (28.9635 + 12.011e-6 * (co2_bkg - 400)) * u.kg / u.mol
    )  # molar mass of dry air [kg/mol]
    return (
        pressure.to(u.Pa)
        * m_a
        / (moist_air_compressibility * R * temperature.to(u.K, equivalencies=u.temperature()))
        * (1 - x_w * (1 - m_w / m_a))
    ).decompose()  # Tomasi eq. 12


# TODO: This function has no unit test
def partial_pressure_water_vapor(temperature, relative_humidity):
    """
    Calculates the partial pressure of water vapor in the air.

    Parameters
    ----------
    temperature : astropy.units.Quantity
        Temperature
    relative_humidity : astropy.units.Quantity
        Relative humidity

    Returns
    -------
    astropy.units.Quantity
        Water vapor partial pressure
    """
    return relative_humidity.to(u.dimensionless_unscaled) * saturation_vapor_pressure(temperature)
